//////////////////////////////////////////////////////////////////
// (c) Copyright 2005-  by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/** @file OptimizableFunctorBase.h
 * @brief Define a base class for one-dimensional functions with optimizable variables
 */
#ifndef QMCPLUSPLUS_OPTIMIZABLEFUNCTORBASE_H
#define QMCPLUSPLUS_OPTIMIZABLEFUNCTORBASE_H
#include <complex>
#include "OhmmsData/libxmldefs.h"
#include "Optimize/VarList.h"

template<class T> struct NumericTraits {};

template<>
struct NumericTraits<double> {
  typedef double          real_type;
  typedef double          value_type;
  typedef std::complex<double> complex_type;
};

template<>
struct NumericTraits<std::complex<double> > {
  typedef double          real_type;
  typedef std::complex<double> value_type;
  typedef std::complex<double> complex_type;
};

/** Base class for any functor used as a source for NumericalJastrow
*/
template<class T>
struct OptimizableFunctorBase: public NumericTraits<T> {

  ///define the real type
  typedef typename NumericTraits<T>::real_type  real_type;
  ///define the value type
  typedef typename NumericTraits<T>::value_type value_type;
  ///define the type of Optimizable Sets
  typedef VarRegistry<real_type>                OptimizableSetType;

  /** evaluate the value at r
   * @param r distance
   *
   * virtual function necessary for a transformation to a numerical functor
   */
  virtual real_type f(real_type r)=0;

  /** evaluate the first derivate 
   * @param r distance
   *
   * virtual function necessary for a transformation to a numerical functor
   */
  virtual real_type df(real_type r)=0;

  /** process xmlnode and registers variables to optimize
   * @param cur xmlNode for a functor
   */
  virtual bool put(xmlNodePtr cur) = 0;

  /** add variables to be optimized
   * @param vlist list to which  derived classes add optimizable variables
   */
  virtual void addOptimizables(OptimizableSetType& vlist) =0;

  /** reset the optimizable variables
   *
   * @param optVariables list of active optimizable variables
   */
  virtual void resetParameters(OptimizableSetType& optVariables)=0;

  /** empty virtual function to help builder classes
  */
  virtual void setDensity(real_type n) { }
};

#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 1756 $   $Date: 2007-02-17 16:18:13 -0600 (Sat, 17 Feb 2007) $
 * $Id: OptimizableFunctorBase.h 1756 2007-02-17 22:18:13Z jnkim $ 
 ***************************************************************************/

